//----------------------------------------------------------------------------
// NAME: KMLImporter
// DESCRIPTION: KMLImporter allows you to import Placemarks from KML and KMZ files
// DEVELOPER: ShockFire
// WEBSITE: http://shockfire.blogger.com
// VERSION: 1.00
//----------------------------------------------------------------------------

using System;
using System.IO;
using System.Net;
using System.Xml;
using System.Drawing;
using System.Threading;
using System.Reflection;
using System.Collections;
using System.Diagnostics;
using System.Globalization;
using System.Windows.Forms;
using System.ComponentModel;
using System.Security.Cryptography;
using ICSharpCode.SharpZipLib.Zip;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using WorldWind;
using WorldWind.Net;
using WorldWind.Renderable;

namespace KMLPlugin
{
	/// <summary>
	/// Main plugin class
	/// </summary>
	public class KMLImporter : WorldWind.PluginEngine.Plugin
	{
		private const string version = "1.00";				// The version of this plugin
		
		private const int IconSizeConstant = 32;			// The default icon size used for scaling

		private RIcons KMLIcons;							// Main Icon container

		private MenuItem tempMenu = new MenuItem();			// Temp menu item for storing file MenuItems
		private MenuItem aboutMenuItem = new MenuItem();	// About menu item
		private MenuItem napalmMenuItem = new MenuItem();	// Napalm enable/disable menu item

		private Hashtable iconStyles = new Hashtable();		// The main Style storage

		private Hashtable bitmapCache = new Hashtable();	// Hashtable to cache Bitmaps from various sources

		internal string KMLPath;							// Temp internal argument passing variable

		private ArrayList networkLinks = new ArrayList();	// Stores created NetworkLinks

		#region Plugin methods
		/// <summary>
		/// Loads this plugin. Initializes variables and adds layers and menu items
		/// </summary>
		public override void Load()
		{
			// Initialize the main Icons layer
			KMLIcons = new RIcons("KML Icons");
			KMLIcons.IsOn = false;

			// Setup Drag&Drop functionality
			m_Application.WorldWindow.DragEnter += new DragEventHandler(WorldWindow_DragEnter);
			m_Application.WorldWindow.DragDrop += new DragEventHandler(WorldWindow_DragDrop);

			// Add a menu item to the File menu and the Help menu
			MenuItem loadMenuItem = new MenuItem();
			loadMenuItem.Text = "Load KML/KMZ file...";
			loadMenuItem.Click += new EventHandler(loadMenu_Click);
			aboutMenuItem.Text = "About KMLImporter";
			aboutMenuItem.Click += new EventHandler(aboutMenu_Click);
			int mergeOrder = 0;
			foreach (MenuItem menuItem in m_Application.MainMenu.MenuItems)
			{
				if (menuItem.Text.Replace("&", "") == "File")
				{
					foreach (MenuItem subMenuItem in menuItem.MenuItems)
					{
						subMenuItem.MergeOrder = mergeOrder;

						if (subMenuItem.Text == "-")
							mergeOrder = 2;		// Everything after this should come after our new items
					}

					tempMenu.Text = menuItem.Text;
					tempMenu.MergeOrder = 1;	// MergeOrder 1 will have 0 before it and 2 after it
					tempMenu.MenuItems.Add(loadMenuItem);
					tempMenu.MenuItems.Add(new MenuItem("-"));
					menuItem.MergeMenu(tempMenu);
				}

				if (menuItem.Text.Replace("&", "") == "Help")
					menuItem.MenuItems.Add(aboutMenuItem);
			}

			// Add a menu item to the Plugins menu
			bool bEnabled = Napalm.NapalmIsEnabled(this.PluginDirectory);
			if (bEnabled)
				napalmMenuItem.Text = "Disable KMLImporter autoupdate";
			else
				napalmMenuItem.Text = "Enable KMLImporter autoupdate";
			napalmMenuItem.Click += new EventHandler(napalmMenu_Click);
			m_Application.PluginsMenu.MenuItems.Add(napalmMenuItem);

			// Some magic to provide backward compability
			Type typecontroller = typeof(MainApplication);
			System.Reflection.PropertyInfo finfo = typecontroller.GetProperty("CmdArgs", BindingFlags.Static|BindingFlags.Public|BindingFlags.GetProperty);
			string[] temp = null;
			if(finfo != null)
			{
				temp = (string[])finfo.GetValue(null, null);

				// If command line arguments are available, try to find one pointing to a kml/kmz file
				if (temp != null)
				{
					foreach (string arg in temp)
					{
						if (!File.Exists(arg))
							continue;

						string fExt = Path.GetExtension(arg);

						if (fExt != ".kml" && fExt != ".kmz")
							continue;

						LoadDiskKM(arg);
						break;
					}
				}
			}

			// Add the main Icons layer to the globe
			m_Application.WorldWindow.CurrentWorld.RenderableObjects.Add(KMLIcons);

			base.Load();
		}

		/// <summary>
		/// Unloads this plugin. Removes layers and menu items
		/// </summary>
		public override void Unload()
		{
			Cleanup();

			m_Application.WorldWindow.CurrentWorld.RenderableObjects.Remove(KMLIcons);

			// Disable Drag&Drop functionality
			this.Application.WorldWindow.DragEnter -= new DragEventHandler(WorldWindow_DragEnter);
			this.Application.WorldWindow.DragDrop -= new DragEventHandler(WorldWindow_DragDrop);

			// Remove the menu items
			foreach (MenuItem menuItem in m_Application.MainMenu.MenuItems)
			{
				if (menuItem.Text.Replace("&", "") == "File")
				{
					foreach (MenuItem subMenuItem in menuItem.MenuItems)
					{
						if (subMenuItem.Text == tempMenu.MenuItems[0].Text)
						{
							menuItem.MenuItems.RemoveAt(subMenuItem.Index+1);
							menuItem.MenuItems.RemoveAt(subMenuItem.Index);
							break;
						}
					}
				}

				if (menuItem.Text.Replace("&", "") == "Help")
					menuItem.MenuItems.Remove(aboutMenuItem);
			}
			tempMenu.MenuItems.Clear();
			m_Application.PluginsMenu.MenuItems.Remove(napalmMenuItem);
			
			try
			{
				// Delete the temp kmz extract directory
				if (Directory.Exists(Path.Combine(this.PluginDirectory, "kmz")))
					Directory.Delete(Path.Combine(this.PluginDirectory, "kmz"), true);

				foreach (string kmlfile in Directory.GetFiles(this.PluginDirectory, "*.kml"))
				{
					try
					{
						File.Delete(kmlfile);
					}
					catch (System.IO.IOException)
					{	}
				}

				foreach (string kmzfile in Directory.GetFiles(this.PluginDirectory, "*.kmz"))
				{
					try
					{
						File.Delete(kmzfile);
					}
					catch (System.IO.IOException)
					{	}
				}
			}
			catch (Exception) {}

			base.Unload();
		}

		#endregion

		#region KMx loading methods
		/// <summary>
		/// Loads either a KML or KMZ file from disk
		/// </summary>
		/// <param name="filename"></param>
		private void LoadDiskKM(string filename)
		{
			if (Path.GetExtension(filename) == ".kmz")
			{
				bool shouldReturn;
				string ExtractedKMLPath = ExtractKMZ(filename, out shouldReturn);
				if (shouldReturn)
					return;
				Spawn_LoadKML(ExtractedKMLPath);
			}
			else
			{
				Spawn_LoadKML(filename);
			}
			KMLIcons.IsOn = true;
		}

		/// <summary>
		/// Loads a KML file in a new thread
		/// </summary>
		/// <param name="path">The path to the KML file to load</param>
		private void Spawn_LoadKML(string path)
		{
			KMLPath = path;

			ThreadStart threadStart = new ThreadStart(LoadKMLFile);
			Thread kmlThread = new System.Threading.Thread(threadStart);
			kmlThread.Name = "KMLImporter worker thread";
			kmlThread.IsBackground = true;
			kmlThread.Start();

			Napalm.Update(this.PluginDirectory, version);
		}

		/// <summary>
		/// Loads a KML file
		/// </summary>
		private void LoadKMLFile()
		{
			if (KMLPath == null || KMLIcons == null)
				return;

			Cleanup();

			WaitMessage waitMessage = new WaitMessage();
			KMLIcons.ChildObjects.Add(waitMessage);

			// Create a reader to read the file
			System.IO.StreamReader sr = new StreamReader(KMLPath);

			// Read all data from the reader
			string kml = sr.ReadToEnd();

			try
			{
				// Load the actual kml data
				LoadKML(kml, KMLIcons);
			}
			catch (Exception ex)
			{
				Utility.Log.Write("KMLImporter: " + ex.ToString());
				MessageBox.Show(
					String.Format(CultureInfo.InvariantCulture, "Error loading KML file '{0}':\n\n{1}", KMLPath, ex.ToString()), 
					"KMLImporter error", 
					MessageBoxButtons.OK,
					MessageBoxIcon.Error,
					MessageBoxDefaultButton.Button1,
					base.Application.RightToLeft == RightToLeft.Yes ? MessageBoxOptions.RtlReading : MessageBoxOptions.ServiceNotification);
			}

			// Close the reader
			sr.Close();

			// Cleanup
			KMLIcons.ChildObjects.Remove(waitMessage);
			KMLPath = null;				
		}

		/// <summary>
		/// Load given KML data
		/// </summary>
		/// <param name="kml">The KML data to load</param>
		internal void LoadKML(string kml, RIcons layer)
		{
			kml = kml.Replace("xmlns=\"http://earth.google.com/kml/2.0\"", "");	// HACK
			kml = kml.Replace("xmlns='http://earth.google.com/kml/2.0'", "");	// HACK

			// Open the downloaded xml in an XmlDocument to allow for XPath searching
			XmlDocument doc = new XmlDocument();
			doc.LoadXml(kml);

			// Try to find some sort of name for this kml from various places
			XmlNode node = doc.SelectSingleNode("//Document[name]/name");
			if (node != null)
				layer.Name = node.InnerText;
			else
			{
				layer.Name = Path.GetFileNameWithoutExtension(KMLPath);
			}

			// Parse Style and StyleMap nodes and store them
			ParseStyles(doc, KMLPath);

			// Load Placemarks recursively and put them in folders
			XmlNode inNode = doc.SelectSingleNode("/kml/Document");
			if (inNode == null)
				inNode = doc.SelectSingleNode("/kml");
			ParsePlacemarks(inNode, layer, KMLPath);
		}

		#endregion

		#region Parsing methods
		/// <summary>
		/// Parses Styles and StyleMaps and stores them
		/// </summary>
		/// <param name="doc">The document to load styles from</param>
		/// <param name="KmlPath">The path to the KML file that is being loaded</param>
		private void ParseStyles(XmlDocument doc, string KmlPath)
		{
			// Load IconStyle elements and extract the images
			XmlNodeList styles = doc.SelectNodes("//Style[IconStyle and @id]");
			foreach (XmlNode style in styles)
			{
				string name = style.Attributes.GetNamedItem("id").InnerText;
				if (iconStyles.ContainsKey(name))
					continue;

				iconStyles.Add(name, GetStyle(style, new Style(), KmlPath));
			}

			// Load StyleMaps and extract the images linked
			XmlNodeList stylemaps = doc.SelectNodes("//StyleMap[@id]");
			foreach (XmlNode stylemap in stylemaps)
			{
				string name = stylemap.Attributes.GetNamedItem("id").InnerText;
				if (iconStyles.ContainsKey(name))
					continue;

				System.Xml.XmlNode stylemapNode = stylemap.SelectSingleNode("Pair[key=\"normal\"]/styleUrl");
				if (stylemapNode == null)
					continue;

				string normalName = stylemapNode.InnerText.Replace("#", "");
				XmlNode normalNode = doc.SelectSingleNode("//Style[@id='"+normalName+"']");

				iconStyles.Add(name, GetStyle(normalNode, new Style(), KmlPath));
			}
		}

		/// <summary>
		/// Parses Placemarks recursively
		/// </summary>
		/// <param name="inNode">The node containing Placemarks or folders</param>
		/// <param name="layer">The layer to add the resulting icons or folders to</param>
		/// <param name="KmlPath">The path to the KML file that is being loaded</param>
		private void ParsePlacemarks(XmlNode inNode, RIcons layer, string KmlPath)
		{
			// Extract and set layer visibility for the current layer
			XmlNode visible = inNode.SelectSingleNode("visibility");
			if (visible != null)
			{
				if (visible.InnerText == "0")
					layer.IsOn = false;
			}

			// Parse all Folders
			ParseFolders(inNode, layer, KmlPath);

			// Parse NetworkLinks
			ParseNetworkLinks(inNode, layer);

			foreach (WorldWind.Renderable.RenderableObject ro in layer.ChildObjects)
			{
				RIcon ricon = ro as RIcon;
				if (ricon != null)
					ricon.HasBeenUpdated = false;
			}

			// Parse all Placemarks that have a name and location
			XmlNodeList placemarks = inNode.SelectNodes("Placemark[name and Point]");
			foreach (XmlNode node in placemarks)
			{
				try
				{
					string name = node.SelectSingleNode("name").InnerText;
					RIcon update = null;

					// Extract the location string for this Placemark node and split it up
					string loc = node.SelectSingleNode("Point/coordinates").InnerText;

					string sLon="0", sLat="0", sAlt="0";
					if (loc.Split(',').Length == 3)				// Includes altitude
					{
						sLon = loc.Substring(0, loc.IndexOf(",")).Trim();
						sLat = loc.Substring(loc.IndexOf(",") + 1, loc.LastIndexOf(",") - loc.IndexOf(",") - 1).Trim();
						sAlt = loc.Substring(loc.LastIndexOf(",") + 1, loc.Length - loc.LastIndexOf(",") - 1).Trim();
					}
					else										// Lat and Lon only (assume 0 altitude)
					{
						sLon = loc.Substring(0, loc.IndexOf(",")).Trim();
						sLat = loc.Substring(loc.LastIndexOf(",") + 1, loc.Length - loc.LastIndexOf(",") - 1).Trim();
					}

					// Convert extracted positions to numbers
					float lat = Convert.ToSingle(sLat, CultureInfo.InvariantCulture);
					float lon = Convert.ToSingle(sLon, CultureInfo.InvariantCulture);
					float alt = Convert.ToSingle(sAlt, CultureInfo.InvariantCulture);

					string desc = null;
					string uri = null;

					// Extract a description and make sure it's not too long
					XmlNode xnode = node.SelectSingleNode("description");
					if (xnode != null)
					{
						string descRaw = xnode.InnerText;
						uri = SearchUri(descRaw);
						desc = RemoveTags(descRaw.Replace("<BR>", "\n").Replace("<br>", "\n").Replace("<br/>", "\n"));
						if (desc.Length > 505)
							desc = desc.Substring(0, 500) + "...";
					}

					float rotation = 0;
					bool bRotated = false;
					string rotRaw = null;

					// Locate a node containing rotation 
					XmlNode rotNode1 = node.SelectSingleNode("Point/rotation");
					if (rotNode1 != null)
						rotRaw = rotNode1.InnerText;
					else 
					{
						XmlNode rotNode2 = node.SelectSingleNode("IconStyle/heading");
						if (rotNode2 != null)
							rotRaw = rotNode2.InnerText;
						else
						{
							XmlNode rotNode3 = node.SelectSingleNode("Style/IconStyle/heading");
							if (rotNode3 != null)
								rotRaw = rotNode3.InnerText;
						}
					}

					// If rotation was found parse it
					if (rotRaw != null)
					{
						rotation = Convert.ToSingle(rotRaw, CultureInfo.InvariantCulture);
						bRotated = true;
					}

					Style style = null;

					// Takes care of 'normal' IconStyles
					XmlNode styleNode = node.SelectSingleNode("styleUrl");
					if (styleNode != null)
					{
						string styleUrl = styleNode.InnerText.Replace("#", "");
						style = (Style)iconStyles[styleUrl];
					}

					// Check if there's in inline Style node
					styleNode = node.SelectSingleNode("Style");
					if (styleNode != null)
					{
						if (style != null)
							style = GetStyle(styleNode, style, KmlPath);
						else
							style = GetStyle(styleNode, new Style(), KmlPath);
					}

					// Check if this icon has to be extruded
					bool bExtrude = false;
					XmlNode extrudeNode = node.SelectSingleNode("Point/extrude");
					if (extrudeNode != null)
					{
						if (extrudeNode.InnerText == "1")
							bExtrude = true;
					}

					// See if this icon already exists, and store it if it does
					foreach (WorldWind.Renderable.RenderableObject ro in layer.ChildObjects)
					{
						RIcon ricon = ro as RIcon;
						if (ricon != null)
						{
							if ((ro.Name == name) && ((style == null) || ((ricon.NormalIcon == style.NormalIcon) && (!ricon.HasBeenUpdated))))
							{
								update = ricon;
								update.HasBeenUpdated = true;
								
								break;
							}
						}
					}

					// If a previous icons has been found update it's location
					if (update != null)
					{
						update.IsRotated = bRotated;
						if (bRotated)
						{
							update.Rotation = Angle.FromDegrees(rotation);
						}
						update.Height = Double.IsNaN(style.NormalScale) ? IconSizeConstant : (int)(style.NormalScale * Math.Min(((Bitmap)bitmapCache[style.NormalIcon]).Height, IconSizeConstant));
						update.Width = Double.IsNaN(style.NormalScale) ? IconSizeConstant : (int)(style.NormalScale * Math.Min(((Bitmap)bitmapCache[style.NormalIcon]).Width, IconSizeConstant));
						update.Description = desc;
						update.SetPosition(lat, lon, alt);
					}
					else
					{
						// Create the icon with either the generated bitmap or the default dot
						if (style != null)
						{
							CreateIcon(layer, name, desc, uri, lat, lon, alt, style, bRotated, rotation, bExtrude);
						}
						else
						{
							// Use the default 'tack' icon if no style was found
							string pal3Path = Path.Combine(this.m_PluginDirectory, "icons/palette-3.png");
							if (File.Exists(pal3Path))
							{
								if (!bitmapCache.Contains(pal3Path))
									bitmapCache.Add(pal3Path, (Bitmap)Bitmap.FromFile(pal3Path));
								Style pinStyle = new Style(GetSubImage(new Style(pal3Path), 448, 64, 64, 64));

								CreateIcon(layer, name, desc, uri, lat, lon, alt, pinStyle, bRotated, rotation, bExtrude);
							}
						}
					}
				}
				catch (Exception ex)
				{	Utility.Log.Write("KMLImporter: " + ex.ToString());		}
			}

			// Cleanup icons that have not been updated
			RemoveUnusedIcons(layer);
			
			// Update metadata for this layer
			layer.MetaData["Child count"] = layer.ChildObjects.Count;
		}

		/// <summary>
		/// Locates Folders and parses them recursively
		/// </summary>
		/// <param name="inNode">The XmlNode to extract the Folders from</param>
		/// <param name="layer">The layer to add the folders to</param>
		/// <param name="KmlPath">The path to the KML file that is being loaded</param>
		private void ParseFolders(XmlNode inNode, RIcons layer, string KmlPath)
		{
			// Find Folders and initialize them recursively
			XmlNodeList folders = inNode.SelectNodes("Folder");
			foreach (XmlNode node in folders)
			{
				try
				{
					// Find the name of the folder
					string foldername = "Folder";
					XmlNode nameNode = node.SelectSingleNode("name");
					if (nameNode != null)
						foldername = nameNode.InnerText;

					// See if the folder already exists
					RIcons folder = null;
					foreach (RenderableObject ro in layer.ChildObjects)
					{
						RIcons ricons = ro as RIcons;
						if ((ricons != null) && (ro.Name == foldername))
						{
							folder = ricons;
						}
					}

					// Create a new folder if it doesn't exist yet
					if (folder == null)
					{
						folder = new RIcons(foldername);
						layer.Add(folder);
					}

					// Parse placemarks into the folder
					ParsePlacemarks(node, folder, KmlPath);
				}
				catch (Exception ex)
				{	Utility.Log.Write("KMLImporter: " + ex.ToString());		}
			}
		}

		/// <summary>
		/// Parse NetworkLinks
		/// </summary>
		/// <param name="doc">The document to load NetworkLinks from</param>
		private void ParseNetworkLinks(XmlNode inNode, RIcons layer)
		{
			// Find NetworkLinks, initialize them and download them for the first time
			XmlNodeList networklinks = inNode.SelectNodes("NetworkLink");
			foreach (XmlNode node in networklinks)
			{
				try
				{
					// Find out the name for this NetworkLink
					string nlName = "NetworkLink";
					XmlNode nameNode = node.SelectSingleNode("name");
					if (nameNode != null)
						nlName = nameNode.InnerText;

					// See if a folder for this NetworkLink already exists
					RIcons folder = null;
					foreach (RenderableObject ro in layer.ChildObjects)
					{
						RIcons ricons = ro as RIcons;
						if ((ricons != null) && (ro.Name == nlName))
						{
							folder = ricons;
						}
					}

					// Create a new folder if none is available
					if (folder == null)
					{
						folder = new RIcons(nlName);
						layer.Add(folder);
					}

					// Find the URL to download the file from
					string loadFile = null;
					XmlNode hrefNode = node.SelectSingleNode("Url/href");
					if ((hrefNode != null) && (hrefNode.InnerText.Length > 0))
						loadFile = hrefNode.InnerText;

					// Give up if no URL can be found
					if (loadFile == null)
						continue;

//					// Initial fire of download
//					try
//					{
//						string saveFile = Path.Combine(this.PluginDirectory, Path.GetFileName(loadFile));
//						WebDownload myClient = new WebDownload(loadFile);
//						myClient.DownloadFile(saveFile);
//
//						string kmlFile = saveFile;
//						if (Path.GetExtension(saveFile) == ".kmz")
//						{
//							bool bError = false;
//							kmlFile = ExtractKMZ(saveFile, out bError);
//
//							if (bError)
//								continue;
//						}
//
//						// Store whatever is in KMLPath and replace it by the file we downloaded
//						string tempPath = KMLPath;
//						KMLPath = kmlFile;
//
//						// Create a reader to read the file
//						System.IO.StreamReader sr = new StreamReader(kmlFile);
//
//						// Read all data from the reader
//						string kml = sr.ReadToEnd();
//
//						// Close stream, no longer needed
//						sr.Close();
//
//						try
//						{
//							// Load the actual kml data
//							LoadKML(kml, folder);
//						}
//						catch (Exception ex)
//						{
//							MessageBox.Show(
//								String.Format(CultureInfo.InvariantCulture, "Error loading KML file '{0}':\n\n{1}", KMLPath, ex.ToString()), 
//								"KMLImporter error", 
//								MessageBoxButtons.OK, 
//								MessageBoxIcon.Error,
//								MessageBoxDefaultButton.Button1,
//								base.Application.RightToLeft == RightToLeft.Yes ? MessageBoxOptions.RtlReading : MessageBoxOptions.ServiceNotification);
//						}
//
//						// Restore KMLPath to whatever it was
//						KMLPath = tempPath;						
//					}
//					catch (Exception ex)
//					{	Utility.Log.Write("KMLImporter: " + ex.ToString());		}

					// Initialize the actual NetworkLink object to handle updates for us
					if ((node.SelectSingleNode("Url/refreshMode") != null) && (node.SelectSingleNode("Url/refreshMode").InnerText == "onInterval"))
					{
						string refreshText = node.SelectSingleNode("Url/refreshInterval").InnerText;
						int refreshSeconds = Convert.ToInt32(refreshText, CultureInfo.InvariantCulture);

						NetworkLink netLink = new NetworkLink(this, folder, loadFile, refreshSeconds*1000);
						netLink.Fire();
						networkLinks.Add(netLink);
					}
				}
				catch (Exception ex)
				{	Utility.Log.Write("KMLImporter: " + ex.ToString());		}
			}
		}
		#endregion

		#region Style handling methods
		/// <summary>
		/// Modifies a style with Style tags loaded from an XmlNode
		/// </summary>
		/// <param name="style">The XmlNode containing override information</param>
		/// <param name="oldStyle">The style to override</param>
		/// <param name="KmlPath">The path to the KML file that is being loaded</param>
		/// <returns>The style with overridden values</returns>
		private Style GetStyle(XmlNode style, Style oldStyle, string KmlPath)
		{
			Style overStyle = oldStyle;

			bool bPalette = false;

			// Determine the scale, if any, for this style
			XmlNode scaleNode = style.SelectSingleNode("IconStyle/scale");
			if (scaleNode != null)
				overStyle.NormalScale = Convert.ToDouble(scaleNode.InnerText, CultureInfo.InvariantCulture);

			// Search for style tags in location 1
			XmlNode hrefNode = style.SelectSingleNode("IconStyle/Icon/href");
			if (hrefNode != null)
			{
				string filename = hrefNode.InnerText;
				if (filename.StartsWith("root://"))													// Use palette bitmap
				{
					filename = Path.Combine(this.m_PluginDirectory, filename.Remove(0, 7));
					if (File.Exists(filename))
					{
						bPalette = true;
						overStyle.NormalIcon = GetDiskImage(filename);
					}
				}
				else if (filename.StartsWith("http://"))											// Use bitmap from the internet
				{
					overStyle.NormalIcon = GetWebImage(filename);
				}
				else if (File.Exists(Path.Combine(Path.GetDirectoryName(KmlPath), filename)))		// Use a file from disk
				{
					overStyle.NormalIcon = GetDiskImage(Path.Combine(Path.GetDirectoryName(KmlPath), filename));
				}
				else if (File.Exists(Path.Combine(this.PluginDirectory, filename)))					// Use a file from disk
				{
					overStyle.NormalIcon = GetDiskImage(Path.Combine(this.PluginDirectory, filename));
				}
			}

			// See if we need to cut this style to a substyle
			XmlNode wNode = style.SelectSingleNode("IconStyle/Icon/w");
			XmlNode hNode = style.SelectSingleNode("IconStyle/Icon/h");
			if (wNode != null && hNode != null)
			{
				int w = Convert.ToInt32(wNode.InnerText, CultureInfo.InvariantCulture);
				int h = Convert.ToInt32(hNode.InnerText, CultureInfo.InvariantCulture);

				int x = 0, y = 0;
				XmlNode xNode = style.SelectSingleNode("IconStyle/Icon/x");
				if (xNode != null)
					x = Convert.ToInt32(xNode.InnerText, CultureInfo.InvariantCulture);
				XmlNode yNode = style.SelectSingleNode("IconStyle/Icon/y");
				if (yNode != null)
					y = Convert.ToInt32(yNode.InnerText, CultureInfo.InvariantCulture);

				if (bPalette)
					overStyle.NormalIcon = GetSubImage(overStyle, x*2, y*2, w*2, h*2);
				else
					overStyle.NormalIcon = GetSubImage(overStyle, x, y, w, h);
			}

			// Search for style tags in a possible secondary location
			XmlNode iconNode = style.SelectSingleNode("icon");
			if (iconNode != null)
			{
				string filename = iconNode.InnerText;
				if (!filename.StartsWith("http://"))
					return null;

				overStyle.NormalIcon = GetWebImage(filename);
			}

			return overStyle;
		}

		/// <summary>
		/// Extracts a rectangular selection from a given style
		/// </summary>
		/// <param name="filename">The file to extract the SubBitmap from</param>
		/// <param name="x">x</param>
		/// <param name="y">y</param>
		/// <param name="w">width</param>
		/// <param name="h">height</param>
		/// <returns>The generated bitmap</returns>
		private string GetSubImage(Style style, int x, int y, int w, int h)
		{
			// Try using a cached version
			string key = style.NormalIcon+ "|" 
				+x.ToString("D5", CultureInfo.InvariantCulture)
				+y.ToString("D5", CultureInfo.InvariantCulture)
				+w.ToString("D5", CultureInfo.InvariantCulture)
				+h.ToString("D5", CultureInfo.InvariantCulture);
			if (bitmapCache.ContainsKey(key))
				return key;

			// Create a new bitmap to draw into
			Bitmap outImage = new Bitmap(w, h);
			Graphics graphics = Graphics.FromImage(outImage);

			// Draw a region into the newly created bitmap
			RectangleF destinationRect = new RectangleF(0, 0, w, h);
			System.Drawing.Bitmap bit = ((Bitmap)bitmapCache[style.NormalIcon]);
			RectangleF sourceRect = new RectangleF(x, bit.Height - y - h, w, h);
			graphics.DrawImage((Bitmap)bitmapCache[style.NormalIcon], destinationRect, sourceRect, GraphicsUnit.Pixel);
			graphics.Flush();

			// Cache the generated bitmap
			bitmapCache.Add(key, outImage);

			return key;
		}

		/// <summary>
		/// Retrieves a bitmap from cache or a file on disk
		/// </summary>
		/// <param name="filename">The filename to open the bitmap from</param>
		/// <returns>The found bitmap</returns>
		private string GetDiskImage(string filename)
		{
			// Try using a cached version
			if (bitmapCache.ContainsKey(filename))
				return filename;

			// Load the bitmap from disk
			Bitmap bit = (Bitmap)Bitmap.FromFile(filename);
			bitmapCache.Add(filename, bit);

			return filename;
		}

		/// <summary>
		/// Retrieves a bitmap from cache or the web
		/// </summary>
		/// <param name="filename">The URI to download the bitmap from</param>
		/// <returns>The downloaded bitmap</returns>
		private string GetWebImage(string filename)
		{
			// Try using a cached version
			if (bitmapCache.ContainsKey(filename))
				return filename;

			// Download the file from the web
			WebDownload myDownload = new WebDownload(filename);
			myDownload.DownloadMemory();

			// Load it into a Bitmap
			Bitmap bit = (Bitmap)System.Drawing.Bitmap.FromStream(myDownload.ContentStream);

			// Cache the downloaded bitmap
			myDownload.Dispose();
			bitmapCache.Add(filename, bit);
			return filename;
		}
		#endregion

		#region Drag&Drop handling methods
		/// <summary>
		/// Checks if the object being dropped is a kml or kmz file
		/// </summary>
		private void WorldWindow_DragEnter(object sender, DragEventArgs e)
		{
			if (DragDropIsValid(e))
				e.Effect = DragDropEffects.All;
		}

		/// <summary>
		/// Handles dropping of a kml/kmz file (by loading that file)
		/// </summary>
		private void WorldWindow_DragDrop(object sender, DragEventArgs e)
		{
			if (DragDropIsValid(e))
			{
				// transfer the filenames to a string array
				string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);

				if (files.Length > 0 && File.Exists(files[0]))
				{
					LoadDiskKM(files[0]);
				}
			}
		}

		/// <summary>
		/// Determines if this plugin can handle the dropped item
		/// </summary>
		private static bool DragDropIsValid(DragEventArgs e)
		{
			if( e.Data.GetDataPresent(DataFormats.FileDrop, false))
			{
				if (((string[])e.Data.GetData(DataFormats.FileDrop)).Length == 1)
				{
					string extension = Path.GetExtension(((string[])e.Data.GetData(DataFormats.FileDrop))[0]).ToLower(CultureInfo.InvariantCulture);
					if ((extension == ".kml") || (extension == ".kmz"))
						return true;
				}
			}
			return false;
		}
		#endregion

		#region Misc. methods
		internal string ExtractKMZ(string filename, out bool bError)
		{
			bError = false;

			// Create a folder 'kmz' to extract the kmz file to
			string ExtractPath = Path.Combine(this.PluginDirectory, "kmz");
			if (!Directory.Exists(ExtractPath))
				Directory.CreateDirectory(ExtractPath);

			// Extract the kmz file
			FastZip fz = new FastZip();
			fz.ExtractZip(filename, ExtractPath, "");

			// Try to find the extracted kml file to load
			string ExtractedKMLPath = null;
			if (File.Exists(Path.Combine(ExtractPath, "doc.kml")))
				ExtractedKMLPath = Path.Combine(ExtractPath, "doc.kml");
			else
			{
				string[] kmlfiles = Directory.GetFiles(ExtractPath, "*.kml");
				if (kmlfiles.Length > 0)
					ExtractedKMLPath = kmlfiles[0];
				else
				{
					MessageBox.Show(
						"The selected KMZ file does not contain a KML file", 
						"KMLImporter error", 
						MessageBoxButtons.OK, 
						MessageBoxIcon.Warning,
						MessageBoxDefaultButton.Button1,
						base.Application.RightToLeft == RightToLeft.Yes ? MessageBoxOptions.RtlReading : MessageBoxOptions.ServiceNotification);

					bError = true;
				}
			}

			return ExtractedKMLPath;
		}

		/// <summary>
		/// Creates a scaled icon on the globe
		/// </summary>
		/// <param name="Name">The name of the item</param>
		/// <param name="Desc">The description</param>
		/// <param name="Lat">The latitude for the icon</param>
		/// <param name="Lon">The longitude for the icon</param>
		/// <param name="Alt">The altitude to draw the icon at</param>
		/// <param name="bitmapPath">The path to the bitmap to show</param>
		private void CreateIcon(RIcons layer, string Name, string Desc, string Uri, float Lat, float Lon, float Alt, 
			Style style, bool bRotated, float rotation, bool bExtrude)
		{
			// Create the icon and set initial settings
			RIcon ic = new RIcon(
				Name,									// name
				Lat,										// latitude
				Lon,										// longitude
				style.NormalIcon,							// helper
				Alt);									// altitude

			// Set optional icon settings
			if (Desc != null)
				ic.Description = Desc;
			if (Uri != null)
				ic.ClickableActionURL = Uri;
			if (bRotated)
			{
				ic.Rotation = Angle.FromDegrees(rotation);
				ic.IsRotated = true;
			}
			ic.m_drawGroundStick = bExtrude;
			ic.Image = (Bitmap)bitmapCache[style.NormalIcon];
			ic.Height = Double.IsNaN(style.NormalScale) ? IconSizeConstant : (int)(style.NormalScale * Math.Min(((Bitmap)bitmapCache[style.NormalIcon]).Height, IconSizeConstant));
			ic.Width = Double.IsNaN(style.NormalScale) ? IconSizeConstant : (int)(style.NormalScale * Math.Min(((Bitmap)bitmapCache[style.NormalIcon]).Width, IconSizeConstant));

			// Add the icon to the layer
			layer.AddIcon(ic);
		}

		/// <summary>
		/// Handles selecting and loading the selected KML/KMZ file
		/// </summary>
		private void loadMenu_Click(object sender, EventArgs e)
		{
			System.Windows.Forms.OpenFileDialog fileDialog = new OpenFileDialog();
			fileDialog.CheckFileExists = true;
			fileDialog.Filter = "KML/KMZ files (*.kml *.kmz)|*.kml;*.kmz";
			fileDialog.Multiselect = false;
			fileDialog.RestoreDirectory = true;
			DialogResult result = fileDialog.ShowDialog();

			if (result == DialogResult.OK)
			{
				LoadDiskKM(fileDialog.FileName);
			}
		}

		/// <summary>
		/// Cleans up used resources
		/// </summary>
		private void Cleanup()
		{
			foreach (RenderableObject ro in KMLIcons.ChildObjects)
			{
				ro.Dispose();
			}
			KMLIcons.ChildObjects.Clear();

			foreach (Object bit in bitmapCache)
			{
				Bitmap bitmap = bit as Bitmap;
				if (bitmap != null)
					bitmap.Dispose();
			}
			bitmapCache.Clear();

			foreach (NetworkLink netLink in networkLinks)
			{
				netLink.Dispose();
			}
			networkLinks.Clear();

			iconStyles.Clear();
		}

		/// <summary>
		/// Shows information about KMLImporter on a Form
		/// </summary>
		private void aboutMenu_Click(object sender, EventArgs e)
		{
			AboutForm aboutForm = new AboutForm();
			aboutForm.ShowDialog();
		}

		/// <summary>
		/// Toggles the Napalm enabled state
		/// </summary>
		private void napalmMenu_Click(object sender, EventArgs e)
		{
			bool bEnabled = Napalm.NapalmChangeStatus(this.PluginDirectory, napalmMenuItem.Text.StartsWith("Enable"));
			if (bEnabled)
				napalmMenuItem.Text = "Disable KMLImporter autoupdate";
			else
				napalmMenuItem.Text = "Enable KMLImporter autoupdate";
		}

		/// <summary>
		/// Returns the URI found in the first href tag
		/// </summary>
		/// <param name="source">The string to search</param>
		/// <returns>The found URI, or null if no URI was found</returns>
		private static string SearchUri(string source)
		{
			int i = source.IndexOf("<a href");
			if (i != -1)
			{
				int start = source.Substring(i).IndexOf("\"") + i+1;
				int end = source.Substring(start+1).IndexOf("\"") + start+1;
				return source.Substring(start, end-start);
			}

			int start2 = source.IndexOf("http://"); 
			if (start2 != -1) 
			{ 
				int end1 = source.Substring(start2+1).IndexOf("\n"); 
				int end2 = source.Substring(start2+1).IndexOf(" "); 
				int end3 = source.Length -1; 
                  
             
				if (end1 == -1) 
					end1 = Int32.MaxValue;
				else
					end1 += start2+1;
				if (end2 == -1) 
					end2 = Int32.MaxValue;
				else
					end2 += start2+1;
				if (end3 == -1) 
					end3 = Int32.MaxValue; 
  
				int compareend1 = (end1 < end2) ? end1 : end2; 
				int compareend2 = (end3 < compareend1) ? end3 : compareend1; 
                  
               
				return source.Substring(start2, compareend2-start2);
			}

			return null;
		}

		/// <summary>
		/// Removes the tags from a string
		/// </summary>
		/// <param name="input">The string to remove the tags from</param>
		/// <returns>The input string without tags</returns>
		private static string RemoveTags(string source)
		{
			while (true)
			{
				if ((source.IndexOf("<") == -1) || (source.IndexOf(">") == -1))
					break;

				int start = source.IndexOf("<");
				int stop = source.IndexOf(">") + 1;
				int count = stop - start;

				if (count < 0)
					break;

				source = source.Remove(start, count);
			}

			source = source.Replace("&nbsp;", "");

			return source.Trim();
		}

		/// <summary>
		/// Removes unused icons from a given layer
		/// </summary>
		/// <param name="layer">The layer to remove icons from</param>
		private static void RemoveUnusedIcons(RIcons layer)
		{
			// Stores removed icons
			ArrayList VFD = new ArrayList();

			// Search for removed icons
			foreach (WorldWind.Renderable.RenderableObject ro in layer.ChildObjects)
			{
				RIcon ricon = ro as RIcon;
				if ((ricon != null) && (!ricon.HasBeenUpdated))
				{
					VFD.Add(ro);
				}
			}

			// Remove all icons that were found to be removed
			foreach (WorldWind.Renderable.RenderableObject ro in VFD)
			{
				layer.Remove(ro);
				ro.Dispose();
			}
		}
		#endregion
	}

	/// <summary>
	/// Napalm autoupdater code
	/// </summary>
	class Napalm
	{
		private const string plugName = "KMLImporter";
		private const string baseUrl = "http://worldwind.arc.nasa.gov";
		private delegate void UpdateDelegate (string PluginDirectory, string version);

		/// <summary>
		/// Empty constructor, because this class only contains static methods
		/// </summary>
		private Napalm()
		{
			
		}

		/// <summary>
		/// Starts an async update
		/// </summary>
		/// <param name="PluginDirectory"></param>
		/// <param name="version"></param>
		internal static void Update(string PluginDirectory, string version)
		{
			UpdateDelegate udel = new UpdateDelegate(WebUpdate);
			udel.BeginInvoke(PluginDirectory, version, null, null);
		}
		/// <summary>
		/// Updates this plugin (and supporting files) from the internet
		/// </summary>
		private static void WebUpdate(string PluginDirectory, string version)
		{
			CultureInfo icy = CultureInfo.InvariantCulture;
			Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;

			// Now go check for a new version (in the background)
			try
			{
				if (!NapalmIsEnabled(PluginDirectory))
					return;

				if (!File.Exists(Path.Combine(PluginDirectory, plugName+".cs")))
					return;

				// Download the versioning file
				string URL = String.Format(icy, "{0}/{1}/{1}_ver.txt", baseUrl, plugName);
				WebClient verDownloader = new WebClient();
				Stream response = new MemoryStream(verDownloader.DownloadData(URL));

				// Create a reader to read the response from the server
				System.IO.StreamReader sr = new StreamReader(response);

				string ver = sr.ReadLine();

				// Try to update if this version appears to be out of date
				if (ver != version)
				{
					try
					{
						if (Convert.ToSingle(ver, CultureInfo.InvariantCulture) < Convert.ToSingle(version, CultureInfo.InvariantCulture))
							return;
					}
					catch (Exception)
					{
						return;
					}

					// Download the new .cs file
					string CsURL = String.Format(icy, "{0}/{1}/{1}.cs", baseUrl, plugName);
					string CsPath = Path.Combine(PluginDirectory, String.Format(icy, "{0}.cs_", plugName));
					WebClient csDownloader = new WebClient();
					csDownloader.DownloadFile(CsURL, CsPath);

					// Napalm v2.0 secure autoupdater block
					try
					{
						// Create a hash of the file that was downloaded
						System.IO.StreamReader streamreader = new StreamReader(CsPath);
						byte[] testStringBytes = GetHashBytes(streamreader.ReadToEnd());

						RSAParameters RSAKeyInfo = new RSAParameters();
						System.IO.StreamReader keyreader = new System.IO.StreamReader(Path.Combine(PluginDirectory, "key"));

						RSAKeyInfo.Modulus = Convert.FromBase64String(keyreader.ReadLine());
						RSAKeyInfo.Exponent = Convert.FromBase64String(keyreader.ReadLine());
						byte[] SignedHashValue = Convert.FromBase64String(sr.ReadLine());

						RSACryptoServiceProvider RSAdecr = new RSACryptoServiceProvider();
						RSAdecr.ImportParameters(RSAKeyInfo);
            
						RSAPKCS1SignatureDeformatter RSADeformatter = new RSAPKCS1SignatureDeformatter(RSAdecr);
						RSADeformatter.SetHashAlgorithm("SHA1");

						if (!RSADeformatter.VerifySignature(testStringBytes, SignedHashValue))
						{
							Utility.Log.Write(String.Format(icy, "{0}: The file signature is not valid!", plugName));
							return;
						}
					}
					catch (Exception ex)		// General signature checking / cryptography error
					{
						Utility.Log.Write("Signature checking error:\n" + ex);
						return;
					}

					// Backup the current file, and replace it with the file that was just downloaded
					if ((File.Exists(CsPath)) && (new FileInfo(CsPath).Length > 2))
					{
						// Just delete a backup of the same version if it already exists
						string bcpPath = Path.Combine(PluginDirectory, plugName+"_v"+version+".cs_");
						if (File.Exists(bcpPath))
							File.Delete(bcpPath);

						string tempPath = Path.Combine(PluginDirectory, plugName+".cs_");
						string plugPath = Path.Combine(PluginDirectory, plugName+".cs");

						// Move the old file out of the way and replace it with the new version
						File.Move(plugPath,	bcpPath);
						for (int i = 0; (i < 5) && (File.Exists(tempPath)); i++)
						{
							try
							{
								File.Move(tempPath,	plugPath);
							}
							catch (Exception)
							{
								System.Threading.Thread.Sleep(800);
							}
						}
					}

					// Notify the user, because we love users
					string message = String.Format(icy, "The {0} plugin has been updated.\n", plugName);
					message += "If you experience any problems it is recommended that you reload the plugin.\n";
					MessageBox.Show(message, 
						plugName+" updated",
						MessageBoxButtons.OK,
						MessageBoxIcon.Information,
						MessageBoxDefaultButton.Button1,
						MessageBoxOptions.ServiceNotification);
				}
				else if (ver != version)		// This means this is probably an internal plugin, so we can't autoupdate it
				{
					if (MessageBox.Show("A new version of the "+plugName+" plugin is available.\nWould you like to go to the website to download the latest version?",
						"PlaneTracker update available",
						MessageBoxButtons.YesNo,
						MessageBoxIcon.Information,
						MessageBoxDefaultButton.Button1,
						MessageBoxOptions.ServiceNotification) == DialogResult.Yes)
						System.Diagnostics.Process.Start("http://www.worldwindcentral.com/wiki/Add-on:KMLImporter");
				}
			}
			catch (Exception)		// We don't really care if this failed
			{ }
		}

		/// <summary>
		/// Gets a byte array representing the hash of a given string
		/// </summary>
		/// <param name="s">The string to hash</param>
		/// <returns>A byte array containing the hash</returns>
		private static byte[] GetHashBytes(string s)
		{
			byte[] data = System.Text.Encoding.UTF8.GetBytes(s);
			byte[] key = Convert.FromBase64String("szLIWrCoPJ3DSWInZx5Ye7sRz0MKBG3JpmgP2KgzlcWGvuJMqNiD77DVQuIRFvgbc5UCEFRhS5Ii5khitfOXhg==");	// A random key that needs to be kept constant
			byte[] hash = new HMACSHA1(key).ComputeHash(data);
			return hash;
		}

		/// <summary>
		/// Checks whether Napalm is enabled
		/// </summary>
		/// <param name="PluginDirectory">The directory that contains the key file</param>
		/// <returns>true if Napalm is enabled, false if Napalm is not enabled</returns>
		internal static bool NapalmIsEnabled(string PluginDirectory)
		{
			string keyPath = Path.Combine(PluginDirectory, "key");
			if (!File.Exists(keyPath))
				return false;

			StreamReader reader = new StreamReader(keyPath);
			string keyline1 = reader.ReadLine();
			string keyline2 = reader.ReadLine();
			string keyline3 = reader.ReadLine();
			reader.Close();

			if ((keyline1 != null) && (keyline2 != null) && (keyline3 != null) && (keyline3.Length > 0))
				return false;
			else
				return true;
		}

		/// <summary>
		/// Sets Napalm's enabled state
		/// </summary>
		/// <param name="PluginDirectory">The directory that contains the key file</param>
		/// <param name="bEnabled">Whether to enable Napalm</param>
		/// <returns>Whether Napalm was enabled</returns>
		internal static bool NapalmChangeStatus(string PluginDirectory, bool bEnabled)
		{
			string keyPath = Path.Combine(PluginDirectory, "key");
			if (!File.Exists(keyPath))
				return false;

			StreamReader reader = new StreamReader(keyPath);
			string keyline1 = reader.ReadLine();
			string keyline2 = reader.ReadLine();
			reader.Close();

			if ((keyline1 == null) || (keyline2 == null))
				return false;

			StreamWriter writer = new StreamWriter(keyPath);
			writer.WriteLine(keyline1);
			writer.WriteLine(keyline2);
			if (!bEnabled)
			{
				string[] possibleText = new string[] {
														"DisableNapalm",
														"You see; random characters",
														"WARNING: Do not try to read the above lines out loud",
														"\"sharks with frickin' laser beams attached to their heads!\"",
														"\"Oh, my, yes.\"",
														"\"Windmills do not work that way! Good night!\"",
														"\"I am Holly, the ship's computer, with an IQ of 6000, the same IQ as 6000 PE teachers.\"",
														"\"Spoon!\"",
														"\"Contrary to popular opinion, cats cannot see in the dark. They just know where you are.\""};
				Random rand = new Random();
				string disabledString = possibleText[rand.Next(possibleText.Length-1)];
				writer.WriteLine(disabledString);
			}
			writer.Flush();
			writer.Close();

			return bEnabled;
		}
	}
	/// <summary>
	/// Represents a NetworkLink. Updates a layer from a remote source periodically.
	/// </summary>
	class NetworkLink
	{
		private string url;
		private RIcons layer;
		private System.Timers.Timer timer = new System.Timers.Timer();
		private KMLImporter owner;
		private bool bUpdating;

		/// <summary>
		/// Creates and initializes a new NetworkLink
		/// </summary>
		/// <param name="owner">The owner of this NetworkLink</param>
		/// <param name="layer">The layer to update the NetworkLink to</param>
		/// <param name="url">The URL to update the NetworkLink from</param>
		/// <param name="interval">The interval, in milliseconds, at which the NetworkLink should update</param>
		internal NetworkLink(KMLImporter owner, RIcons layer, string url, int interval)
		{
			this.owner = owner;
			this.url = url;
			this.layer = layer;
			timer.Interval = (double)interval;
			timer.Elapsed += new System.Timers.ElapsedEventHandler(timer_Elapsed);
			timer.Start();
		}

		/// <summary>
		/// Fires of a download
		/// </summary>
		internal void Fire()
		{
			timer_Elapsed(null, null);
		}

		/// <summary>
		/// Downloads a KML/KMZ file from the given URL
		/// </summary>
		private void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
		{
			if (bUpdating)
				return;
			bUpdating = true;

			try
			{
				if ((layer != null) && layer.IsOn)
				{
					string saveFile = Path.Combine(owner.PluginDirectory, Path.GetFileName(url));
					WebDownload myClient = new WebDownload(url);
					myClient.DownloadFile(saveFile);

					// Extract the file if it is a kmz file
					string kmlFile = saveFile;
					if (Path.GetExtension(saveFile) == ".kmz")
					{
						bool bError = false;
						kmlFile = owner.ExtractKMZ(saveFile, out bError);

						if (bError)
							return;
					}

					owner.KMLPath = kmlFile;

					// Create a reader to read the file
					StreamReader sr = new StreamReader(kmlFile);

					// Read all data from the reader
					string kml = sr.ReadToEnd();

					sr.Close();

					try
					{
						// Load the actual kml data
						owner.LoadKML(kml, layer);
					}
					catch (Exception ex)
					{
						Utility.Log.Write("KMLImporter: " + ex.ToString());

						MessageBox.Show(
							String.Format(CultureInfo.InvariantCulture, "Error loading KML file '{0}':\n\n{1}", kmlFile, ex.ToString()), 
							"KMLImporter error", 
							MessageBoxButtons.OK, 
							MessageBoxIcon.Error,
							MessageBoxDefaultButton.Button1,
							MessageBoxOptions.ServiceNotification);
					}
				}
			}
			catch (Exception ex)
			{
				Utility.Log.Write("KMLImporter: " + ex.ToString());
			}

			bUpdating = false;
		}

		/// <summary>
		/// Stops the timer
		/// </summary>
		internal void Dispose()
		{
			timer.Stop();
		}
	}

	/// <summary>
	/// Helper class. Represents a KML Style or StyleMap
	/// </summary>
	class Style
	{
		#region Public members
		internal string NormalIcon
		{
			get
			{
				return normalIcon;
			}
			set
			{
				normalIcon = value;
			}
		}

		internal double NormalScale
		{
			get
			{
				return normalScale;
			}
			set
			{
				normalScale = value;
			}
		}

		#endregion

		#region Private members
		private string normalIcon;
		private double normalScale = Double.NaN;
		#endregion

		/// <summary>
		/// Creates a new Style
		/// </summary>
		internal Style()
		{
			this.normalIcon = null;
		}

		/// <summary>
		/// Creates a new Style
		/// </summary>
		/// <param name="normalIcon">The normal Bitmap to use for this Style</param>
		internal Style(string normalIcon)
		{
			this.normalIcon = normalIcon;
		}
	}

	/// <summary>
	/// Holds a collection of icons
	/// </summary>
	class RIcons : RenderableObjectList
	{
		#region Private members
		private Hashtable m_textures = new Hashtable(); // Texture cache
		
		private Sprite m_sprite;
		private Line m_groundStick;
		private Vector2[] groundStick = new Vector2[2];

		private RIcon mouseOverIcon; // The closest icon the mouse is currently over

		private const int minIconZoomAltitude = 5000000;

		private static int hotColor = Color.White.ToArgb();
		private static int normalColor = Color.FromArgb(150,255,255,255).ToArgb();
		private static int descriptionColor = Color.White.ToArgb();
		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref= "T:RIcons"/> class 
		/// </summary>
		/// <param name="name">The name to give the layer</param>
		internal RIcons(string name) : base(name) 
		{
			this.MetaData.Add("Child count", 0);
		}

		/// <summary>
		/// Adds an icon to this layer. Deprecated.
		/// </summary>
		internal void AddIcon(RIcon icon)
		{
			Add(icon);
		}

		/// <summary>
		/// Draw the icon
		/// </summary>
		protected virtual void Render(DrawArgs drawArgs, RIcon icon, Vector3 projectedPoint)
		{
			if (!icon.isInitialized)
				icon.Initialize(drawArgs);

			if(!drawArgs.WorldCamera.ViewFrustum.ContainsPoint(icon.Position))
				return;

			IconTexture iconTexture = GetTexture(icon);
			bool isMouseOver = icon == mouseOverIcon;
			if(isMouseOver)
			{
				// Mouse is over
				isMouseOver = true;

				if(icon.isSelectable)
					DrawArgs.MouseCursor = CursorType.Hand;

				string description = icon.Description;
				if(description==null)
					description = icon.ClickableActionURL;
				if(description!=null)
				{
					// Render description field
					DrawTextFormat format = DrawTextFormat.NoClip | DrawTextFormat.WordBreak | DrawTextFormat.Bottom;
					int left = 10;
					if(World.Settings.ShowLayerManager)
						left += World.Settings.LayerManagerWidth;
					Rectangle rect = Rectangle.FromLTRB(left, 10, drawArgs.screenWidth - 10, drawArgs.screenHeight - 10 );

					// Draw outline
					drawArgs.defaultDrawingFont.DrawText(
						m_sprite, description,
						rect,
						format, 0xb0 << 24 );
					
					rect.Offset(2,0);
					drawArgs.defaultDrawingFont.DrawText(
						m_sprite, description,
						rect,
						format, 0xb0 << 24 );

					rect.Offset(0,2);
					drawArgs.defaultDrawingFont.DrawText(
						m_sprite, description,
						rect,
						format, 0xb0 << 24 );

					rect.Offset(-2,0);
					drawArgs.defaultDrawingFont.DrawText(
						m_sprite, description,
						rect,
						format, 0xb0 << 24 );

					// Draw description
					rect.Offset(1,-1);
					drawArgs.defaultDrawingFont.DrawText(
						m_sprite, description,
						rect, 
						format, descriptionColor );
				}
			}

			int color = isMouseOver ? hotColor : normalColor;
			if(iconTexture==null || isMouseOver)
			{
				// Render label
				if(icon.Name != null)
				{
					// Render name field
					const int labelWidth = 1000; // Dummy value needed for centering the text
					if(iconTexture==null)
					{
						// Center over target as we have no bitmap
						Rectangle rect = new Rectangle(
							(int)projectedPoint.X - (labelWidth>>1), 
							(int)(projectedPoint.Y - (drawArgs.defaultDrawingFont.Description.Height >> 1)),
							labelWidth, 
							drawArgs.screenHeight );

						drawArgs.defaultDrawingFont.DrawText(m_sprite, icon.Name, rect, DrawTextFormat.Center, color);
					}
					else
					{
						// Adjust text to make room for icon
						int spacing = (int)(icon.Width * 0.3f);
						if(spacing>10)
							spacing = 10;
						int offsetForIcon = (icon.Width>>1) + spacing;

						Rectangle rect = new Rectangle(
							(int)projectedPoint.X + offsetForIcon, 
							(int)(projectedPoint.Y - (drawArgs.defaultDrawingFont.Description.Height >> 1)),
							labelWidth, 
							drawArgs.screenHeight );

						drawArgs.defaultDrawingFont.DrawText(m_sprite, icon.Name, rect, DrawTextFormat.WordBreak, color);
					}
				}
			}

			if (icon.m_drawGroundStick)
			{
				Vector3 projectedGroundPoint = drawArgs.WorldCamera.Project(icon.m_groundPoint);

				m_groundStick.Begin();
				groundStick[0].X = projectedPoint.X;
				groundStick[0].Y = projectedPoint.Y;
				groundStick[1].X = projectedGroundPoint.X;
				groundStick[1].Y = projectedGroundPoint.Y;

				m_groundStick.Draw(groundStick, Color.Red.ToArgb());
				m_groundStick.End();
			}

			if(iconTexture!=null)
			{
				float factor = 1;
				if (drawArgs.WorldCamera.Altitude > minIconZoomAltitude)
					factor -= (float)((drawArgs.WorldCamera.Altitude - minIconZoomAltitude) / drawArgs.WorldCamera.Altitude);

				// Render icon
				float xscale = factor * ((float)icon.Width / iconTexture.Width);
				float yscale = factor * ((float)icon.Height / iconTexture.Height);
				m_sprite.Transform = Matrix.Scaling(xscale, yscale, 0);
				
				if (icon.IsRotated) m_sprite.Transform *= Matrix.RotationZ((float)icon.Rotation.Radians - (float) drawArgs.WorldCamera.Heading.Radians);

				m_sprite.Transform *= Matrix.Translation(projectedPoint.X, projectedPoint.Y, 0);
				m_sprite.Draw( iconTexture.Texture,
					new Vector3(iconTexture.Width>>1, iconTexture.Height>>1,0),
					Vector3.Empty,
					color );
				
				// Reset transform to prepare for text rendering later
				m_sprite.Transform = Matrix.Identity;
			}
		}

		/// <summary>
		/// Retrieve an icon's texture
		/// </summary>
		private IconTexture GetTexture(RIcon icon)
		{
			object key = icon.Image;
			if(key==null)
				return null;
			IconTexture res = (IconTexture)m_textures[key];
			return res;
		}


		#region RenderableObject methods
		/// <summary>
		/// Add a child object to this layer.
		/// </summary>
		public override void Add(RenderableObject ro)
		{
			m_children.Add(ro);
			isInitialized = false;
		}

		public override void Initialize(DrawArgs drawArgs)
		{
			if(!isOn)
				return;

			if(m_sprite != null)
			{
				m_sprite.Dispose();
				m_sprite = null;
			}
			if(m_groundStick != null)
			{
				m_groundStick.Dispose();
				m_groundStick = null;
			}

			m_sprite = new Sprite(drawArgs.device);
			m_groundStick = new Line(drawArgs.device);

			// Load all textures
			foreach(RenderableObject ro in m_children)
			{
				RIcon icon = ro as RIcon;
				if(icon==null)
				{
					// Child is not an icon
					if(ro.IsOn)
						ro.Initialize(drawArgs);
					continue;
				}

				// Child is an icon
				icon.Initialize(drawArgs);

				object key = null;
				IconTexture iconTexture = null;

				// Icon image from bitmap
				if(icon.Image != null)
				{
					iconTexture = (IconTexture)m_textures[icon.Image];
					if(iconTexture==null)
					{
						// Create new texture from image
						key = icon.Image;
						iconTexture = new IconTexture( drawArgs.device, icon.Image);
					}
				}

				if(iconTexture==null)
					// No texture set
					continue;

				if(key!=null)
				{
					// New texture, cache it
					m_textures.Add(key, iconTexture);

					// Use default dimensions if not set
					if(icon.Width==0)
						icon.Width = iconTexture.Width;
					if(icon.Height==0)
						icon.Height = iconTexture.Height;
				}
			}

			// Compute mouse over bounding boxes
			foreach(RenderableObject ro in m_children)
			{
				RIcon icon = ro as RIcon;
				if(icon==null)
					// Child is not an icon
					continue;

				if(GetTexture(icon)==null)
				{
					// Label only 
					icon.SelectionRectangle = drawArgs.defaultDrawingFont.MeasureString(null, icon.Name, DrawTextFormat.None, 0);
				}
				else
				{
					// Icon only
					icon.SelectionRectangle = new Rectangle(0, 0, icon.Width, icon.Height);
				}

				// Center the box at (0,0)
				icon.SelectionRectangle.Offset(-icon.SelectionRectangle.Width/2, -icon.SelectionRectangle.Height/2 );
			}

			isInitialized = true;
		}

		public override void Dispose()
		{
			base.Dispose();

			if(m_textures != null)
			{
				foreach(IconTexture iconTexture in m_textures.Values)
					iconTexture.Texture.Dispose();
				m_textures.Clear();
			}

			if(m_sprite != null)
			{
				m_sprite.Dispose();
				m_sprite = null;
			}

			if (m_groundStick != null)
			{
				m_groundStick.Dispose();
				m_groundStick = null;
			}
		}

		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			foreach(RenderableObject ro in m_children)
			{
				if(!ro.IsOn)
					continue;
				if(!ro.isSelectable)
					continue;

				RIcon icon = ro as RIcon;
				if(icon==null)
				{
					// Child is not an icon
					if (ro.PerformSelectionAction(drawArgs))
						return true;
					continue;
				}

				if(!drawArgs.WorldCamera.ViewFrustum.ContainsPoint(icon.Position))
					continue;

				Vector3 projectedPoint = drawArgs.WorldCamera.Project(icon.Position);
				if(!icon.SelectionRectangle.Contains(
					DrawArgs.LastMousePosition.X - (int)projectedPoint.X, 
					DrawArgs.LastMousePosition.Y - (int)projectedPoint.Y ) )
					continue;

				try
				{
					Process.Start(icon.ClickableActionURL);
					return true;
				}
				catch
				{
				}
			}
			return false;
		}

		public override void Render(DrawArgs drawArgs)
		{
			if(!isOn)
				return;

			if(!isInitialized)
				return;

			// First render everything except icons
			foreach(RenderableObject ro in m_children)
			{
				if(ro is RIcon)
					continue;

				if(!ro.IsOn)
					continue;

				// Child is not an icon
				ro.Render(drawArgs);
			}

			int closestIconDistanceSquared = int.MaxValue;
			RIcon closestIcon = null;

			// Now render just the icons
			m_sprite.Begin(SpriteFlags.AlphaBlend);
			foreach(RenderableObject ro in m_children)
			{
				if(!ro.IsOn)
					continue;


				RIcon icon = ro as RIcon;
				if(icon==null)
					continue;

				// Find closest mouse-over icon
				Vector3 projectedPoint = drawArgs.WorldCamera.Project(icon.Position);

				int dx = DrawArgs.LastMousePosition.X - (int)projectedPoint.X;
				int dy = DrawArgs.LastMousePosition.Y - (int)projectedPoint.Y;
				if( icon.SelectionRectangle.Contains( dx, dy ) )
				{
					// Mouse is over, check whether this icon is closest
					int distanceSquared = dx*dx + dy*dy;
					if(distanceSquared < closestIconDistanceSquared)
					{
						closestIconDistanceSquared = distanceSquared;
						closestIcon = icon;
					}
				}

				if(icon != mouseOverIcon)
					Render(drawArgs, icon, projectedPoint);
			}

			// Render the mouse over icon last (on top)
			if(mouseOverIcon != null)
				Render(drawArgs, mouseOverIcon, drawArgs.WorldCamera.Project(mouseOverIcon.Position));

			mouseOverIcon = closestIcon;

			m_sprite.End();
		}
		#endregion
	}

	/// <summary>
	/// One icon in an RIcons layer
	/// </summary>
	class RIcon : RenderableObject
	{
		#region Public members
		internal string Description;			// Longer description of icon (addition to name)
		internal Bitmap Image;					// Icon image
		internal int Width;						// Icon on-screen rendered width (pixels)
		internal int Height;					// Icon on-screen rendered height (pixels)
		internal string ClickableActionURL // On-Click browse to location
		{
			get
			{
				return m_clickableActionURL;
			}
			set 
			{
				isSelectable = value != null;
				m_clickableActionURL = value;
			}
		}
		internal Rectangle SelectionRectangle;	// Bounding box centered at (0,0) used to calculate whether mouse is over icon/label

		internal Angle Rotation
		{
			get
			{
				return m_rotation;
			}
			set
			{
				m_rotation = value;
			}
		}

		internal bool IsRotated
		{
			get
			{
				return m_isRotated;
			}
			set
			{
				m_isRotated = value;
			}
		}


		internal bool m_drawGroundStick = true;
		internal Vector3 m_groundPoint = new Vector3();

		internal string NormalIcon;

		internal bool HasBeenUpdated = true;
		#endregion

		#region Private members
		private Angle m_rotation;
		private bool m_isRotated;

		private string m_clickableActionURL;

		private float m_latitude;
		private float m_longitude;
		private float m_altitude;				// The icon altitude above sea level
		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref= "T:RIcon"/> class 
		/// </summary>
		/// <param name="name">Name of the icon</param>
		/// <param name="latitude">Latitude in decimal degrees.</param>
		/// <param name="longitude">Longitude in decimal degrees.</param>
		/// <param name="heightAboveSurface">Icon height (meters) above sea level.</param>
		internal RIcon(string name, float latitude, float longitude, string normalicon, float heightAboveSurface) : base( name )
		{
			m_latitude = latitude;
			m_longitude = longitude;
			NormalIcon = normalicon;
			this.m_altitude = heightAboveSurface;
		}

		/// <summary>
		/// Sets the geographic position of the icon.
		/// </summary>
		/// <param name="latitude">Latitude in decimal degrees.</param>
		/// <param name="longitude">Longitude in decimal degrees.</param>
		/// <param name="altitude">The icon altitude above sea level.</param>
		internal void SetPosition(float latitude, float longitude, float altitude)
		{
			m_latitude = latitude;
			m_longitude = longitude;
			this.m_altitude = altitude;

			// Recalculate XYZ coordinates
			isInitialized = false;
		}


		#region RenderableObject methods
		public override void Initialize(DrawArgs drawArgs)
		{
			Position = MathEngine.SphericalToCartesian(m_latitude, m_longitude, 
				m_altitude + drawArgs.WorldCamera.WorldRadius);

			if (m_drawGroundStick)
			{
				// compute the ground point
				double elevation = drawArgs.WorldCamera.WorldRadius;

				if((((WorldWindow)drawArgs.parentControl).CurrentWorld.TerrainAccessor != null) && (drawArgs.WorldCamera.Altitude < 300000))
				{
					elevation += ((WorldWindow)drawArgs.parentControl).CurrentWorld.TerrainAccessor.GetElevationAt(
						m_latitude,
						m_longitude,
						100.0) * World.Settings.VerticalExaggeration;
				}

				m_groundPoint = MathEngine.SphericalToCartesian(m_latitude, m_longitude, elevation);
			}

			isInitialized = true;
		}

		/// <summary>
		/// Disposes the icon (when disabled)
		/// </summary>
		public override void Dispose()
		{
			// Nothing to dispose
		}

		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			// Handled by parent
			return false;
		}

		public override void Update(DrawArgs drawArgs)
		{
			// Handled by parent
		}

		public override void Render(DrawArgs drawArgs)
		{
			// Handled by parent
		}
		#endregion
	}

	/// <summary>
	/// Renders a message to the lower right corner
	/// </summary>
	class WaitMessage : RenderableObject
	{
		#region Private members
		private string _Text = "Please wait, loading KML file.";
		private int color = Color.White.ToArgb();
		private int distanceFromCorner = 25;
		#endregion

		/// <summary>
		/// Creates a new WaitMessage
		/// </summary>
		internal WaitMessage() : base("KML WaitMessage", Vector3.Empty, Quaternion.Identity)
		{
			// We want to be drawn on top of everything else
			this.RenderPriority = RenderPriority.Icons;

			// true to make this layer active on startup, this is equal to the checked state in layer manager
			this.IsOn = true;
		}


		#region RenderableObject methods
		/// <summary>
		/// This is where we do our rendering 
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override void Render(DrawArgs drawArgs)
		{
			// Draw the current text using default font in lower right corner
			Rectangle bounds = drawArgs.defaultDrawingFont.MeasureString(null, _Text, DrawTextFormat.None, 0);
			drawArgs.defaultDrawingFont.DrawText(null, _Text, 
				drawArgs.screenWidth-bounds.Width-distanceFromCorner, drawArgs.screenHeight-bounds.Height-distanceFromCorner,
				color );
		}

		/// <summary>
		/// RenderableObject abstract member (needed) 
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Initialize(DrawArgs drawArgs)
		{
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Update(DrawArgs drawArgs)
		{
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Dispose()
		{
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			return false;
		}
		#endregion
	}

	/// <summary>
	/// A Form with information about KMLImporter
	/// </summary>
	class AboutForm : System.Windows.Forms.Form
	{
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.Label label2;
		private System.Windows.Forms.LinkLabel linkLabel1;
		private System.Windows.Forms.Button button1;
		private System.Windows.Forms.Label label3;
		private System.ComponentModel.Container components = null;

		internal AboutForm()
		{
			InitializeComponent();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}


		private void button1_Click(object sender, EventArgs e)
		{
			this.Close();
		}

		private void linkLabel1_Click(object sender, EventArgs e)
		{
			try
			{
				System.Diagnostics.Process.Start("http://shockfire.blogspot.com/");
			}
			catch (Exception) {}
		}


		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.label1 = new System.Windows.Forms.Label();
			this.label2 = new System.Windows.Forms.Label();
			this.linkLabel1 = new System.Windows.Forms.LinkLabel();
			this.button1 = new System.Windows.Forms.Button();
			this.label3 = new System.Windows.Forms.Label();
			this.SuspendLayout();
			// 
			// label1
			// 
			this.label1.Font = new System.Drawing.Font("Microsoft Sans Serif", 12F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.label1.Location = new System.Drawing.Point(120, 8);
			this.label1.Name = "label1";
			this.label1.Size = new System.Drawing.Size(160, 48);
			this.label1.TabIndex = 0;
			this.label1.Text = "KMLImporter";
			this.label1.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// label2
			// 
			this.label2.Location = new System.Drawing.Point(96, 144);
			this.label2.Name = "label2";
			this.label2.Size = new System.Drawing.Size(224, 24);
			this.label2.TabIndex = 1;
			this.label2.Text = "Created by Tim van den Hamer (ShockFire)";
			this.label2.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// linkLabel1
			// 
			this.linkLabel1.Location = new System.Drawing.Point(96, 176);
			this.linkLabel1.Name = "linkLabel1";
			this.linkLabel1.Size = new System.Drawing.Size(224, 16);
			this.linkLabel1.TabIndex = 2;
			this.linkLabel1.TabStop = true;
			this.linkLabel1.Text = "http://shockfire.blogspot.com/";
			this.linkLabel1.Click += new EventHandler(linkLabel1_Click);
			this.linkLabel1.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// button1
			// 
			this.button1.Location = new System.Drawing.Point(168, 216);
			this.button1.Name = "button1";
			this.button1.TabIndex = 3;
			this.button1.Text = "OK";
			this.button1.Click += new EventHandler(button1_Click);
			// 
			// label3
			// 
			this.label3.Location = new System.Drawing.Point(96, 64);
			this.label3.Name = "label3";
			this.label3.Size = new System.Drawing.Size(224, 64);
			this.label3.TabIndex = 4;
			this.label3.Text = "KMLImporter - A placemark importer for NASA World Wind allowing easy importing of kml/kmz files.";
			// 
			// AboutForm
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(416, 265);
			this.ControlBox = false;
			this.Controls.Add(this.label3);
			this.Controls.Add(this.button1);
			this.Controls.Add(this.linkLabel1);
			this.Controls.Add(this.label2);
			this.Controls.Add(this.label1);
			this.ShowInTaskbar = false;
			this.MaximizeBox = false;
			this.MinimizeBox = false;
			this.Name = "AboutForm";
			this.Text = "About KMLImporter";
			this.ResumeLayout(false);
		}
		#endregion
	}
}